#include <stdlib.h>
#include <math.h>

#include "multiscatter.h"

#define PI 3.14159265358979323846

static ms_real *msj_bscat = NULL;
static ms_real *msj_x = NULL;

/* Number of gates currently allocated in the intermediate
   variables: */
static int msj_num_gates_allocated = 0;

/* Number of gates defined in the last call to the algorithm: */
static int msj_num_gates_defined = 0;


static
int
msj_init_intermediate_arrays(int num_gates)
{
  /* Macro for allocating or reallocating memory for a single array */
#define ALLOCATE(target, num_bytes) \
   if ((target = realloc(target, num_bytes)) == NULL) { \
      target = NULL; return MS_FAILURE; }

  int i;
  if (msj_num_gates_allocated < num_gates) {
    /* Number of gates allocated is less than number requested (which
       may be zero if this function has not been called): (re)allocate
       memory. */
    int num_bytes = num_gates * sizeof(ms_real);
    msj_num_gates_allocated = 0; /* In case error occurs */
    ALLOCATE(msj_bscat, num_bytes);
    ALLOCATE(msj_x, num_bytes);
  }
  msj_num_gates_allocated = num_gates;
  /* Set each element of each array to zero. */
  for (i = 0; i < num_gates; i++) {
    msj_bscat[i] = 0.0;
    msj_x[i] = 1.0;
  }
  msj_num_gates_defined = num_gates;
  return MS_SUCCESS;
}

static
int
multiscatter_with_numerical_jacobian(
    /* Input data */
    int n,                    /* Number of range gates in profile */
    ms_real wavelength,       /* Wavelength of laser, microns */
    ms_real rho_laser,        /* 1/e laser beam divergence, radians */
    ms_real rho_telescope,    /* Half-angle field of view, radians */
    ms_real lidar_altitude,   /* Lidar altitude, metres */
    ms_real *range,           /* Height of each range gate, metres */
    ms_real *radius,          /* Cloud/aerosol equivalent radius, microns */
    ms_real *ext,             /* Cloud/aerosol extinction coefficient, m-1 */
    ms_real *ext_bscat_ratio, /* Cloud/aerosol ext./backscatter ratio, sr */
    ms_real *ext_air,         /* Air ext. coefft, m-1 (NULL for vacuum) */
    int n_x,  /* Number of state variables to compute the Jacobian of */
    int n_y,  /* Number of measurements to compute the Jacobian of */
    int *i_x, /* Index to the required state variables */
    int *i_y, /* Index to the required measurements */
    /* Output data: each Jacobian may be NULL if not required */
    ms_real *bscat,               /* Measured backscatter, m-1 sr-1 */
    ms_real **d_ln_beta_d_ln_ext,    /* Jacobian with respect to ln(ext) */
    ms_real **d_ln_beta_d_ln_radius, /* Jacobian w.r.t. ln(radius) */
    ms_real *d_ln_beta_d_ln_ext_bscat_ratio /* ...w.r.t. ln(ext_bscat_ratio) */
    )
{
  ms_real tweak = 100;
  ms_real exp_tweak = exp(1.0/tweak);
  /* Ordinary run */
  multiscatter(n, wavelength, rho_laser, rho_telescope,
	       lidar_altitude, range, radius,
	       ext, ext_bscat_ratio, ext_air,
	       bscat);
  msj_init_intermediate_arrays(n);

  if (d_ln_beta_d_ln_ext_bscat_ratio) {
    int l;
    for (l = 0; l < n_y; l++) {
      if (ext[i_y[l]] > 0.0) {
	d_ln_beta_d_ln_ext_bscat_ratio[l] = -(ms_bscat_single[i_y[l]]
		      + ms_bscat_double[i_y[l]] + ms_bscat_multi[i_y[l]])
	  / bscat[i_y[l]];
      }
      else {
	d_ln_beta_d_ln_ext_bscat_ratio[l] = 0.0;
      }
    }
    /*
    int k;
    for (k = 0; k < n_x; k++) {
      d_ln_beta_d_ln_ext_bscat_ratio[k] = -(ms_bscat_single[i_x[k]]
		 + ms_bscat_double[i_x[k]] + ms_bscat_multi[i_x[k]])/bscat[i_x[k]];
    }
    */
  }
  if (d_ln_beta_d_ln_ext) {
    /* Copy ext into x */
    int i, k, l;
    for (i = 0; i < n; i++) {
      msj_x[i] = ext[i];
    }
    /* Loop through all the x values to find the derivative for */
    for (k = 0; k < n_x; k++) {
      msj_x[i_x[k]] *= exp_tweak;
      multiscatter(n, wavelength, rho_laser, rho_telescope,
		   lidar_altitude, range, radius,
		   msj_x, ext_bscat_ratio, ext_air,
		   msj_bscat);
      for (l = 0; l < n_y; l++) {
	d_ln_beta_d_ln_ext[l][k] = tweak * (msj_bscat[i_y[l]]/bscat[i_y[l]] - 1.0);
      }
      msj_x[i_x[k]] = ext[i_x[k]];
    }
  }
  if (d_ln_beta_d_ln_radius) {
    /* Copy radius into x */
    int i, k, l;
    for (i = 0; i < n; i++) {
      msj_x[i] = radius[i];
    }
    /* Loop through all the x values to find the derivative for */
    for (k = 0; k < n_x; k++) {
      msj_x[i_x[k]] *= exp_tweak;
      multiscatter(n, wavelength, rho_laser, rho_telescope,
		   lidar_altitude, range, msj_x,
		   ext, ext_bscat_ratio, ext_air,
		   msj_bscat);
      for (l = 0; l < n_y; l++) {
	d_ln_beta_d_ln_radius[l][k] = tweak * (msj_bscat[i_y[l]]/bscat[i_y[l]] - 1.0);
      }
      msj_x[i_x[k]] = radius[i_x[k]];
    }
  }
  return MS_SUCCESS;
}

static
int
multiscatter_with_analytic_jacobian(
    /* Input data */
    int n,                    /* Number of range gates in profile */
    ms_real wavelength,       /* Wavelength of laser, microns */
    ms_real rho_laser,        /* 1/e laser beam divergence, radians */
    ms_real rho_telescope,    /* Half-angle field of view, radians */
    ms_real lidar_altitude,   /* Lidar altitude, metres */
    ms_real *range,           /* Height of each range gate, metres */
    ms_real *radius,          /* Cloud/aerosol equivalent radius, microns */
    ms_real *ext,             /* Cloud/aerosol extinction coefficient, m-1 */
    ms_real *ext_bscat_ratio, /* Cloud/aerosol ext./backscatter ratio, sr */
    ms_real *ext_air,         /* Air ext. coefft, m-1 (NULL for vacuum) */
    int n_x,  /* Number of state variables to compute the Jacobian of */
    int n_y,  /* Number of measurements to compute the Jacobian of */
    int *i_x, /* Index to the required state variables */
    int *i_y, /* Index to the required measurements */
    /* Output data */
    ms_real *bscat,               /* Measured backscatter, m-1 sr-1 */
    ms_real **d_ln_beta_d_ln_ext,    /* Jacobian with respect to ln(ext) */
    ms_real **d_ln_beta_d_ln_radius, /* Jacobian w.r.t. ln(radius) */
    ms_real *d_ln_beta_d_ln_ext_bscat_ratio /* ...w.r.t. ln(ext_bscat_ratio) */
    )
{
  /* Variance of laser */
  ms_real mu2_laser = rho_laser*rho_laser;        /* radians^2 */

  /* Variance of telescope: this is a top-hat distribution */
  ms_real mu2_telescope = 0.5 * rho_telescope*rho_telescope;  /* radians^2 */
  ms_real fov_factor = 1.0 / (1.0 - exp(-rho_telescope*rho_telescope
				     / (rho_laser*rho_laser)));
  int range_sign_factor = range[2] < range[1] ? -1 : 1;

  ms_real self_extinction_factor = 0.5;
  ms_real self_scatter_factor = 0.5;

  ms_real drange = fabs(range[2]-range[1]);

  if (ms_options & MS_CRUDE_INTEGRATION) {
    self_scatter_factor = 1.0;
  }
  else if (ms_options & MS_NO_MULTISCAT_WITHIN_GATE) {
    self_scatter_factor = 0.0;
  }

  if (ms_options & MS_CRUDE_OPTICAL_DEPTH) {
    self_extinction_factor = 1.0;
  }

  msj_init_intermediate_arrays(n);

  multiscatter(n, wavelength, rho_laser, rho_telescope,
	       lidar_altitude, range, radius,
	       ext, ext_bscat_ratio, ext_air,
	       bscat);

  if (d_ln_beta_d_ln_ext_bscat_ratio) {
    int l;
    for (l = 0; l < n_y; l++) {
      if (ext[i_y[l]] > 0.0) {
	d_ln_beta_d_ln_ext_bscat_ratio[l] = -(ms_bscat_single[i_y[l]]
		      + ms_bscat_double[i_y[l]] + ms_bscat_multi[i_y[l]])
	  / bscat[i_y[l]];
      }
      else {
	d_ln_beta_d_ln_ext_bscat_ratio[l] = 0.0;
      }
    }
  }
  if (d_ln_beta_d_ln_ext && d_ln_beta_d_ln_radius) {
    int k, l;
    /* Loop through all the x values to find the derivative for */
    for (k = 0; k < n_x; k++) {
      ms_real theta = wavelength / (PI*radius[i_x[k]]);
      ms_real theta2 = theta*theta;
      for (l = 0; l < n_y; l++) {
	  ms_real total_range = range_sign_factor 
	    * (range[i_y[l]]-lidar_altitude);
	  ms_real total_range2 = total_range * total_range;
	  ms_real delta_range = range_sign_factor 
	    * (range[i_y[l]]-range[i_x[k]]);
	  ms_real delta_range2 = delta_range * delta_range;
	/* Single scattering */
	d_ln_beta_d_ln_radius[l][k] = 0.0;

	if (i_x[k] > i_y[l]) {
	  /* Observation pixel closer to lidar than state pixel: no
	     effect of state on observation and Jacobian = 0 */
	  d_ln_beta_d_ln_ext[l][k] = 0.0;
	}
	else if (i_x[k] < i_y[l]) {
	  ms_real denominator = theta2*delta_range2/total_range2+mu2_laser;
	  ms_real exp_term = exp(-2.0*mu2_telescope / denominator);
	  /* Observation pixel is further from lidar than state pixel:
	     effect due to extinction and multiple scattering */
	  d_ln_beta_d_ln_ext[l][k] =
	    /* extinction: */
	    -2.0 * ext[i_x[k]] * drange
	    /* double scattering: */
	    + (ms_bscat_single[i_y[l]] + ms_bscat_air_single[i_y[l]])
	    * fov_factor * drange * ext[i_x[k]]
	    * (1 - exp_term) / bscat[i_y[l]];
	  /* Radius term due to double scattering */
	  d_ln_beta_d_ln_radius[l][k] = 
	    (ms_bscat_single[i_y[l]] + ms_bscat_air_single[i_y[l]])
	    * fov_factor * drange * ext[i_x[k]] / bscat[i_y[l]]
	    * exp_term
	    * 2.0*mu2_telescope/(denominator*denominator)
	    * 2.0*theta2*delta_range2/total_range2;
	}
	else {
	  /* Observation and state pixel are the same: effect due to
	     extinction, multiple scattering and the ext/bscat
	     ratio */ 
	  d_ln_beta_d_ln_ext[l][k] = (
	    /* Single scattering: */
	    ms_bscat_single[i_x[k]] + ms_bscat_double[i_x[k]]
	                            + ms_bscat_multi[i_x[k]]
	    -self_extinction_factor * 2.0 * ext[i_x[k]] 
	                            * bscat[i_y[l]] * drange
	    /* Double scattering */
	    +self_scatter_factor * (ms_bscat_single[i_y[l]]
				    + ms_bscat_air_single[i_y[l]])
	    * fov_factor * fov_factor * drange * ext[i_x[k]]) / bscat[i_y[l]];
	}
      }
    }
  }
  return MS_SUCCESS;
}

int
multiscatter_with_jacobian(
    /* Input data */
    int n,                    /* Number of range gates in profile */
    ms_real wavelength,       /* Wavelength of laser, microns */
    ms_real rho_laser,        /* 1/e laser beam divergence, radians */
    ms_real rho_telescope,    /* Half-angle field of view, radians */
    ms_real lidar_altitude,   /* Lidar altitude, metres */
    ms_real *range,           /* Height of each range gate, metres */
    ms_real *radius,          /* Cloud/aerosol equivalent radius, microns */
    ms_real *ext,             /* Cloud/aerosol extinction coefficient, m-1 */
    ms_real *ext_bscat_ratio, /* Cloud/aerosol ext./backscatter ratio, sr */
    ms_real *ext_air,         /* Air ext. coefft, m-1 (NULL for vacuum) */
    int n_x,  /* Number of state variables to compute the Jacobian of */
    int n_y,  /* Number of measurements to compute the Jacobian of */
    int *i_x, /* Index to the required state variables */
    int *i_y, /* Index to the required measurements */
    /* Output data */
    ms_real *bscat_out,           /* Measured backscatter, m-1 sr-1 */
    ms_real **d_ln_beta_d_ln_ext,    /* Jacobian with respect to ln(ext) */
    ms_real **d_ln_beta_d_ln_radius, /* Jacobian w.r.t. ln(radius) */
    ms_real *d_ln_beta_d_ln_ext_bscat_ratio /* ...w.r.t. ln(ext_bscat_ratio) */
    )
{
  int status;
  if (ms_options & MS_NUMERICAL_JACOBIAN) {
    status = multiscatter_with_numerical_jacobian(
	   n, wavelength, rho_laser, rho_telescope, lidar_altitude, 
	   range, radius, ext, ext_bscat_ratio, ext_air,
	   n_x, n_y, i_x, i_y, 
	   bscat_out,
	   d_ln_beta_d_ln_ext, d_ln_beta_d_ln_radius, 
	   d_ln_beta_d_ln_ext_bscat_ratio);
  }
  else {
    status = multiscatter_with_analytic_jacobian(
	   n, wavelength, rho_laser, rho_telescope, lidar_altitude, 
	   range, radius, ext, ext_bscat_ratio, ext_air,
	   n_x, n_y, i_x, i_y, 
	   bscat_out,
	   d_ln_beta_d_ln_ext, d_ln_beta_d_ln_radius, 
	   d_ln_beta_d_ln_ext_bscat_ratio);
  }
  return status;
}
